from traitlets import Unicode

from .fileprovider import TraefikFileProviderProxy


class TraefikTomlProxy(TraefikFileProviderProxy):
    """Deprecated alias for file provider"""

    toml_dynamic_config_file = Unicode(
        config=True,
    ).tag(
        deprecated_in="1.0",
        deprecated_for="TraefikFileProvider.dynamic_config_file",
    )

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.log.warning(
            "TraefikTomlProxy is deprecated in jupyterhub-traefik-proxy 1.0. Use `c.JupyterHub.proxy_class = 'traefik_file'"
        )



"""Tests for the authentication to the traefik proxy api (dashboard)"""
import pytest
from jupyterhub.utils import exponential_backoff
from tornado.httpclient import AsyncHTTPClient

# Mark all tests in this file as asyncio
pytestmark = pytest.mark.asyncio


@pytest.mark.parametrize(
    "username, password, expected_rc",
    [("api_admin", "admin", 200), ("api_admin", "1234", 401), ("", "", 401)],
)
async def test_traefik_api_auth(proxy, username, password, expected_rc):
    traefik_api_url = proxy.traefik_api_url + "/api/overview"

    async def api_login():
        try:
            if not username and not password:
                resp = await AsyncHTTPClient().fetch(traefik_api_url)
            else:
                resp = await AsyncHTTPClient().fetch(
                    traefik_api_url,
                    auth_username=username,
                    auth_password=password,
                )
        except ConnectionRefusedError:
            rc = None
        except Exception as e:
            rc = e.response.code
        else:
            rc = resp.code
        return rc

    async def cmp_api_login():
        rc = await api_login()
        if rc == expected_rc:
            return True
        else:
            print(f"{rc} != {expected_rc}")
            return False

    await exponential_backoff(cmp_api_login, "Traefik API not reacheable")

    rc = await api_login()
    assert rc == expected_rc
    return



import json
import os

import pytest

from jupyterhub_traefik_proxy import traefik_utils


# Mark all tests in this file as asyncio
def test_roundtrip_routes():
    pytest.mark.asyncio
    routes = {
        "backends": {
            "backend1": {
                "servers": {"server1": {"url": "http://127.0.0.1:9009", "weight": 1}}
            }
        },
        "frontends": {
            "frontend1": {
                "backend": "backend1",
                "passHostHeader": True,
                "routes": {
                    "test": {
                        "rule": "Host:host;PathPrefix:/proxy/path",
                        "data": json.dumps({"test": "test1"}),
                    }
                },
            }
        },
    }

    file_name = "test_roudtrip.toml"
    config_handler = traefik_utils.TraefikConfigFileHandler(file_name)
    config_handler.atomic_dump(routes)
    reloader = traefik_utils.TraefikConfigFileHandler(file_name)
    reloaded = reloader.load()
    os.remove(file_name)
    assert reloaded == routes


def test_atomic_writing(tmpdir):
    testfile = tmpdir.join("testfile")
    with testfile.open("w") as f:
        f.write("before")

    with traefik_utils.atomic_writing(str(testfile)) as f:
        f.write("after")

    # tempfile got cleaned up
    assert not os.path.exists(f.name)

    # file was updated
    with testfile.open("r") as f:
        assert f.read() == "after"

    # didn't leave any residue
    assert tmpdir.listdir() == [testfile]


def test_atomic_writing_recovery(tmpdir):
    testfile = tmpdir.join("testfile")
    with testfile.open("w") as f:
        f.write("before")

    with pytest.raises(TypeError):
        with traefik_utils.atomic_writing(str(testfile)) as f:
            f.write("after")
            f.write(b"invalid")

    # tempfile got cleaned up
    assert not os.path.exists(f.name)

    # file didn't get overwritten
    with testfile.open("r") as f:
        assert f.read() == "before"

    # didn't leave any residue
    assert tmpdir.listdir() == [testfile]



import argparse
import asyncio
import logging
import sys
import textwrap
import time
from contextlib import asynccontextmanager, contextmanager, nullcontext
from inspect import isawaitable
from multiprocessing import cpu_count
from pathlib import Path
from subprocess import Popen
from tempfile import TemporaryDirectory

import aiohttp
import numpy as np
from jupyterhub.proxy import ConfigurableHTTPProxy
from jupyterhub.tests.mocking import MockHub
from jupyterhub.utils import wait_for_http_server

from jupyterhub_traefik_proxy.consul import TraefikConsulProxy
from jupyterhub_traefik_proxy.etcd import TraefikEtcdProxy
from jupyterhub_traefik_proxy.fileprovider import TraefikFileProviderProxy

aiohttp.TCPConnector.__init__.__kwdefaults__['limit'] = 10

performance_dir = Path(__file__).parent.resolve()


def configure_argument_parser():
    parser = argparse.ArgumentParser(
        description="Performance measurement utility",
        epilog=textwrap.dedent(
            """\
            Available measurements:
            - Proxy methods:
                - add_route
                - delete_route
                - get_all_routes
            - Request throughput: small requests
            - Request throughput: large requests
            - Websocket throughput
            """
        ),
        formatter_class=argparse.RawTextHelpFormatter,
    )

    parser.add_argument(
        "metric",
        nargs="?",
        default="methods",
        choices=[
            "methods",
            "http_throughput_small",
            "http_throughput_large",
            "ws_throughput_small",
            "ws_throughput_large",
        ],
        help=textwrap.dedent(
            """\
            What metric to measure. Available metrics:
            - methods
            - http_throughput_small
            - http_throughput_large
            - ws_throughput
            If no metric is provided, it defaults to:
            --- %(default)s ---
            """
        ),
    )

    parser.add_argument(
        "--proxy",
        dest="proxy_class",
        default="file",
        help=textwrap.dedent(
            """\
            Proxy class to analyze.
            Available proxies:
            - file
            - etcd
            - consul
            - chp
            If no proxy is provided, it defaults to:
            --- %(default)s ---
            """
        ),
    )

    parser.add_argument(
        "--routes",
        dest="routes",
        default=10,
        help=textwrap.dedent(
            """\
            Number of routes to be added/deleted/retrieved.
            If no number is provided, it defaults to:
            --- %(default)s ---
            """
        ),
    )

    parser.add_argument(
        "-j",
        "--concurrency",
        dest="concurrency",
        default=10,
        help=textwrap.dedent(
            """\
            Number of concurrent requests when computing the throughput.
            Number of concurrent API calls when computing method performance.
            If no number is provided, it defaults to:
            --- %(default)s ---
            """
        ),
    )

    parser.add_argument(
        "--requests",
        dest="total_requests",
        default=1000,
        help=textwrap.dedent(
            """\
            Total number of requests when computing the throughput.
            If no number is provided, it defaults to:
            --- %(default)s ---
            """
        ),
    )

    parser.add_argument(
        "--iterations",
        dest="test_iterations",
        default=1,
        help=textwrap.dedent(
            """
            How many times to run the measurement.
            If no value is provided, it defaults to:
            --- %(default)s ---
            """
        ),
    )

    parser.add_argument(
        "--output",
        dest="csv_filename",
        help=textwrap.dedent(
            """
            The csv file name where the results will be stored.
            If no file is provided, the results will only be outputed to stdout.
            """
        ),
    )

    return parser


@contextmanager
def etcd():
    """Context manager for running etcd"""
    with TemporaryDirectory() as td:
        p = Popen(['etcd'], cwd=td)
        # wait for it to start
        time.sleep(5)
        try:
            yield
        finally:
            try:
                p.terminate()
            except Exception as e:
                print(f"Error stopping {p}: {e}", file=sys.stderr)


@contextmanager
def consul():
    """Context manager for running consul"""
    with TemporaryDirectory() as td:
        p = Popen(['consul', 'agent', '-dev'], cwd=td)
        # wait for it to start
        time.sleep(5)
        try:
            yield
        finally:
            try:
                p.terminate()
            except Exception as e:
                print(f"Error stopping {p}: {e}", file=sys.stderr)


@asynccontextmanager
async def backend(concurrency=4):
    port = 9000
    # limit backend workers to 1 per CPU
    concurrency = min(concurrency, cpu_count())

    p = Popen(
        [
            "uvicorn",
            "dummy_http_server:app",
            f"--workers={concurrency}",
            f"--port={port}",
            "--log-level=warning",
        ],
        cwd=performance_dir,
    )
    try:
        await wait_for_http_server(f"http://127.0.0.1:{port}")
        yield port
    finally:
        try:
            p.terminate()
        except Exception as e:
            print(f"Error stopping {p}: {e}", file=sys.stderr)


@contextmanager
def measure_time(print_message, stdout_print, time_taken):
    real_time = time.perf_counter()
    process_time = time.process_time()
    yield
    real_time = time.perf_counter() - real_time
    cpu_time = time.process_time() - process_time
    io_time = max(0.0, real_time - cpu_time)
    if stdout_print:
        print(f"{print_message}")
        print(f"CPU time:      {cpu_time:.3f} s")
        print(f"IO time:       {io_time:.3f} s")
        print(f"REAL time:     {real_time:.3f} s")
        # print()
    time_taken["cpu"] = cpu_time
    time_taken["real"] = real_time


async def no_auth_consul_proxy():
    """
    Function returning a configured TraefikEtcdProxy.
    No etcd authentication.
    """

    proxy = TraefikConsulProxy(
        public_url="http://127.0.0.1:8000",
        traefik_api_password="admin",
        traefik_api_username="admin",
        should_start=True,
        # traefik_log_level="DEBUG",
    )
    await proxy.start()
    return proxy


async def no_auth_etcd_proxy():
    """
    Function returning a configured TraefikEtcdProxy.
    No etcd authentication.
    """
    proxy = TraefikEtcdProxy(
        public_url="http://127.0.0.1:8000",
        traefik_api_password="admin",
        traefik_api_username="admin",
        should_start=True,
    )
    await proxy.start()
    return proxy


async def file_proxy():
    """Function returning a configured TraefikFileProviderProxy"""
    proxy = TraefikFileProviderProxy(
        public_url="http://127.0.0.1:8000",
        traefik_api_password="admin",
        traefik_api_username="admin",
        should_start=True,
        # traefik_log_level="DEBUG",
    )

    await proxy.start()
    return proxy


async def configurable_http_proxy():
    """Function returning a configured ConfigurableHTTPProxy"""
    proxy = ConfigurableHTTPProxy(
        auth_token="secret!",
        api_url="http://127.0.0.1:54321",
        should_start=True,
        public_url="http://127.0.0.1:8000",
    )

    app = MockHub()
    app.init_hub()
    proxy.app = app
    proxy.hub = app.hub

    await proxy.start()
    return proxy


@asynccontextmanager
async def get_proxy(proxy_class):
    logging.basicConfig(level=logging.INFO)
    parent_context = nullcontext
    if proxy_class == "file":
        proxy_f = file_proxy
    elif proxy_class == "etcd":
        proxy_f = no_auth_etcd_proxy
        parent_context = etcd
    elif proxy_class == "consul":
        proxy_f = no_auth_consul_proxy
        parent_context = consul
    elif proxy_class == "chp":
        proxy_f = configurable_http_proxy
    else:
        raise ValueError(f"Proxy {proxy_class} not supported")
        return

    with parent_context():
        proxy = await proxy_f()
        try:
            yield proxy
        finally:
            stop = proxy.stop()
            if isawaitable(stop):
                await stop
    # let everything cleanup
    await asyncio.sleep(3)


def format_method_result(
    method,
    test_id,
    sample,
    results,
    const_fields,
):
    result = {}
    result.update(const_fields)
    result["method"] = method
    result["test_id"] = test_id
    result["route_idx"] = sample
    result["cpu_time"] = results[test_id][method][sample]["cpu"]
    result["real_time"] = results[test_id][method][sample]["real"]
    return result


def persist_methods_results_to_csv(
    csv_writer,
    results,
    test_iterations,
    samples,
    const_fields,
):
    for test_id in range(test_iterations):
        for sample in results[test_id]["add"].keys():
            result_add_dict = format_method_result(
                "add", test_id, sample, results, const_fields
            )
            result_delete_dict = format_method_result(
                "delete", test_id, sample, results, const_fields
            )

            csv_writer.writerow(result_add_dict)
            csv_writer.writerow(result_delete_dict)
            if sample in results[test_id]["get_all"]:
                result_get_all_dict = format_method_result(
                    "get_all",
                    test_id,
                    sample,
                    results,
                    const_fields,
                )
                csv_writer.writerow(result_get_all_dict)


def logspace_samples(routes):
    sample_no = 3
    if routes > 40:
        sample_no = routes // 10

    samples = np.unique(
        np.logspace(0, np.log10(routes), sample_no, endpoint=False, dtype=int)
    )
    return samples



"""Traefik implementation

Custom proxy implementations can subclass :class:`Proxy`
and register in JupyterHub config:

.. sourcecode:: python

    from mymodule import MyProxy
    c.JupyterHub.proxy_class = MyProxy

Route Specification:

- A routespec is a URL prefix ([host]/path/), e.g.
  'host.tld/path/' for host-based routing or '/path/' for default routing.
- Route paths should be normalized to always start and end with '/'
"""

# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

import asyncio
from collections.abc import Mapping
from functools import wraps
from numbers import Number

from traitlets import Unicode

from . import traefik_utils
from .proxy import TraefikProxy


def _one_at_a_time(method):
    """decorator to limit an async method to be called only once

    If multiple concurrent calls to this method are made,
    piggy-back on the outstanding call instead of queuing
    or letting requests pile up.
    """

    @wraps(method)
    async def locked_method(*args, **kwargs):
        if getattr(method, "_shared_future", None) is not None:
            f = method._shared_future
            if f.done():
                method._shared_future = None
            else:
                return await f

        method._shared_future = f = asyncio.ensure_future(method(*args, **kwargs))
        try:
            return await f
        finally:
            method._shared_future = None

    return locked_method


class TKvProxy(TraefikProxy):
    """
    JupyterHub Proxy implementation using traefik and a key-value store.

    Custom proxy implementations based on traefik and a key-value store
    can sublass :class:`TKvProxy`.
    """

    kv_traefik_prefix = traefik_utils.KVStorePrefix(
        "traefik",
        config=True,
        help="""The key value store key prefix for traefik static configuration""",
    )

    kv_jupyterhub_prefix = traefik_utils.KVStorePrefix(
        "jupyterhub",
        config=True,
        help="""The key value store key prefix for traefik dynamic configuration""",
    )

    kv_separator = Unicode(
        "/",
        config=True,
        help="""The separator used for the path in the KV store""",
    )

    # these should be the only three methods a KV provider needs to define

    async def _kv_atomic_set(self, to_set: dict):
        """Set a collection of keys and values

        Should be done atomically (i.e. in a transaction),
        setting nothing on failure.

        Args:

        to_set (dict): key/value pairs to set
            Will always be a flattened dict
            of single key-value pairs,
            not a nested structure.
        """
        raise NotImplementedError()

    async def _kv_atomic_delete(self, *keys):
        """Delete one or more keys

        If a key ends with `self.kv_separator`, it should be a recursive delete
        """
        raise NotImplementedError()

    async def _kv_get_tree(self, prefix):
        """Return all data under prefix as a dict

        Should probably use `unflatten_dict_from_kv`
        """
        raise NotImplementedError()

    # now: implement methods required by TraefikProxy base class

    async def _apply_dynamic_config(self, dynamic_config, jupyterhub_config=None):
        """Apply dynamic config (and optional jupyterhub info) atomically"""
        to_set = self.flatten_dict_for_kv(dynamic_config, prefix=self.kv_traefik_prefix)
        if jupyterhub_config:
            to_set.update(
                self.flatten_dict_for_kv(
                    jupyterhub_config, prefix=self.kv_jupyterhub_prefix
                )
            )
        self.log.debug("Setting key-value config %s", to_set)
        await self._kv_atomic_set(to_set)

    async def _delete_dynamic_config(self, traefik_keys, jupyterhub_keys):
        """Delete keys from dynamic configuration

        Translate key paths to flat kv keys
        """
        to_delete = [
            self.kv_separator.join([self.kv_traefik_prefix] + key_path + [""])
            for key_path in traefik_keys
        ]
        to_delete.extend(
            self.kv_separator.join([self.kv_jupyterhub_prefix] + key_path + [""])
            for key_path in jupyterhub_keys
        )
        async with self.semaphore:
            try:
                await self._kv_atomic_delete(*to_delete)
            except Exception as e:
                self.log.error("Couldn't delete config %s: %s", to_delete, e)
                raise

    @_one_at_a_time
    async def _get_jupyterhub_dynamic_config(self):
        """jupyterhub data is in our kv store"""
        return await self._kv_get_tree(self.kv_jupyterhub_prefix)

    async def get_route(self, routespec):
        """Return the route info for a given routespec.

        Args:
            routespec (str):
                A URI that was used to add this route,
                e.g. `host.tld/path/`

        Returns:
            result (dict):
                dict with the following keys::

                'routespec': The normalized route specification passed in to add_route
                    ([host]/path/)
                'target': The target host for this route (proto://host)
                'data': The arbitrary data dict that was passed in by JupyterHub when adding this
                        route.

            None: if there are no routes matching the given routespec
        """
        routespec = self.validate_routespec(routespec)
        router_alias = traefik_utils.generate_alias(routespec, "router")
        route_key = self.kv_separator.join(
            [self.kv_jupyterhub_prefix, "routes", router_alias]
        )
        route = await self._kv_get_tree(route_key)
        if not route:
            return None
        return {key: route[key] for key in ("routespec", "data", "target")}

    # deep/flat dict translation

    def _kv_to_str(self, value):
        """KV stores only store strings

        serialize numbers and booleans to strings that traefik will recognize
        """
        if isinstance(value, str):
            return value
        elif isinstance(value, bool):
            return str(value).lower()
        elif isinstance(value, Number):
            return str(value)
        else:
            # don't coerce unrecognized types, TypeError will be raised later
            return value

    def flatten_dict_for_kv(self, data, prefix=''):
        """Flatten a dictionary of `data` for storage in the KV store,
        prefixing each key with `prefix` and joining each key with
        :attr:`TKvProxy.kv_separator`.

        If the final value is a :class:`list`, then the provided bottom-level key
        shall be appended with an incrementing numeric number, in the style
        that is used by traefik's KV store, e.g.

        .. code-block::

            flatten_dict_for_kv({
                'x' : {
                    'y' : {
                        'z': 'a'
                    }
                }, {
                    'foo': 'bar'
                },
                'baz': [ 'a', 'b', 'c' ]
            })

        :return: The flattened dictionary
        :rtype: dict

        e.g.

        .. code-block::

            {
                 'traefik/x/y/z' : 'a',
                 'traefik/x/foo': 'bar',
                 'traefik/baz/0': 'a',
                 'traefik/baz/1': 'b',
                 'traefik/baz/2': 'c',
            }

        Inspired by `this answer on StackOverflow <https://stackoverflow.com/a/6027615>`_
        """
        sep = self.kv_separator
        # if it's a list, cast to dict of str keys, i.e. ["x"] -> {"0": "x"}
        # if isinstance(data, list):
        #     data = {str(i): item for i, item in enumerate(data)}

        items = {}
        for k, v in data.items():
            if prefix:
                new_key = f"{prefix}{sep}{k}"
            else:
                new_key = k

            if isinstance(v, list):
                # if it's a list, cast to dict of str keys, i.e. ["x"] -> {"0": "x"}
                v = {str(i): item for i, item in enumerate(v)}

            # two cases:
            # 1. Mapping (dict)
            # 2. scalar (cast to str)
            if isinstance(v, Mapping):
                if not v:
                    self.log.warning(
                        f"Not setting anything for empty dict at {new_key}"
                    )
                items.update(self.flatten_dict_for_kv(v, prefix=new_key))
            else:
                # cast _known_ types to str
                v = self._kv_to_str(v)
                # if cast didn't coerce, we can't handle it
                if not isinstance(v, str):
                    raise ValueError(
                        f"Cannot upload {new_key}: {v} of type {type(v)} to kv store"
                    )
                items[new_key] = v
        return items

    def unflatten_dict_from_kv(self, kv_list, root_key=""):
        """Reconstruct tree dict from list of key/value pairs

        This is the inverse of flatten_dict_for_kv,
        not including str coercion.

        Args:

        kv_list (list):
            list of (key, value) pairs.
            keys and values should all be strings.
        root_key (str, optional):
            The key representing the root of the tree,
            if not the root of the key-value store.

        Returns:

        tree (dict):
            The reconstructed dictionary.
            All values will still be strings,
            even those that originated as numbers or booleans.
        """

        sep = self.kv_separator

        def by_depth(item):
            """sort-key function to get shallow items first

            Ensures

            So that we know that we always see an item
            before its children
            """
            key, value = item
            key_path = key.split(sep)
            for i, label in enumerate(key_path):
                # ensure
                if label.isdigit():
                    key_path[i] = int(label)
            return len(key.split(sep))

        tree = {}
        for key, value in sorted(kv_list, key=by_depth):
            key_path = key.split(sep)
            d = tree
            for parent_key, key in zip(key_path[:-1], key_path[1:]):
                if parent_key not in d:
                    # create container
                    if key.isdigit():
                        # integer keys mean it's a list
                        d[parent_key] = []
                    else:
                        d[parent_key] = {}
                # walk down to the next level
                d = d[parent_key]
            if isinstance(d, list):
                # validate list keys
                if len(d) != int(key):
                    raise IndexError(
                        f"Got invalid list key {key_path}, missing previous items in {d}"
                    )
                d.append(value)
            else:
                d[key] = value

        if root_key:
            original_tree = tree
            # get the root of the tree,
            # rather than starting from root
            for key in root_key.strip(self.kv_separator).split(self.kv_separator):
                if key not in tree:
                    self.log.warning(
                        f"Root key {root_key!r} not found in {original_tree}"
                    )
                    return {}
                tree = tree[key]
        return tree



import os
import subprocess
import sys

import pytest

installer_module = "jupyterhub_traefik_proxy.install"

# Mark all tests in this file as slow
pytestmark = pytest.mark.slow


def assert_only_traefik_existence(deps_dir):
    assert deps_dir.exists()
    assert os.listdir(deps_dir) == ["traefik"]


def test_default_conf(tmp_path):
    default_deps_dir = tmp_path / "dependencies"

    subprocess.run([sys.executable, "-m", installer_module], cwd=str(tmp_path))
    assert_only_traefik_existence(default_deps_dir)


def test_output_arg_new_dir(tmp_path):
    deps_dir = tmp_path / "deps" / "out"
    subprocess.run(
        [sys.executable, "-m", installer_module, "--traefik", f"--output={deps_dir}"]
    )
    assert_only_traefik_existence(deps_dir)


def test_output_arg_existing_dir(tmp_path):
    deps_dir = tmp_path / "deps"
    deps_dir.mkdir()
    subprocess.run(
        [sys.executable, "-m", installer_module, "--traefik", f"--output={deps_dir}"]
    )
    assert_only_traefik_existence(deps_dir)


def test_version(tmp_path):
    deps_dir = tmp_path / "deps"
    subprocess.run(
        [
            sys.executable,
            "-m",
            installer_module,
            f"--output={deps_dir}",
            "--traefik",
            "--traefik-version=2.8.8",
        ]
    )
    assert_only_traefik_existence(deps_dir)


@pytest.mark.parametrize(
    "platform",
    [
        "linux-amd64",
        "linux-arm64",
        "darwin-arm64",
        "darwin-amd64",
    ],
)
def test_platform(tmp_path, platform):
    deps_dir = tmp_path / "deps"
    subprocess.run(
        [
            sys.executable,
            "-m",
            installer_module,
            f"--output={deps_dir}",
            f"--platform={platform}",
        ]
    )
    assert_only_traefik_existence(deps_dir)



"""Traefik implementation

Custom proxy implementations can subclass :class:`Proxy`
and register in JupyterHub config:

.. sourcecode:: python

    from mymodule import MyProxy
    c.JupyterHub.proxy_class = MyProxy

Route Specification:

- A routespec is a URL prefix ([host]/path/), e.g.
  'host.tld/path/' for host-based routing or '/path/' for default routing.
- Route paths should be normalized to always start and end with '/'
"""

# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

from concurrent.futures import ThreadPoolExecutor
from urllib.parse import urlparse

from tornado.concurrent import run_on_executor
from traitlets import Any, Bool, List, Unicode, default

from .kv_proxy import TKvProxy


class TraefikEtcdProxy(TKvProxy):
    """JupyterHub Proxy implementation using traefik and etcd"""

    executor = Any()

    provider_name = "etcd"

    etcd_client_ca_cert = Unicode(
        config=True,
        allow_none=True,
        default_value=None,
        help="""Etcd client root certificates""",
    )

    etcd_client_cert_crt = Unicode(
        config=True,
        allow_none=True,
        default_value=None,
        help="""Etcd client certificate chain
            (etcd_client_cert_key must also be specified)""",
    )

    etcd_client_cert_key = Unicode(
        config=True,
        allow_none=True,
        default_value=None,
        help="""Etcd client private key
            (etcd_client_cert_crt must also be specified)""",
    )

    # The grpc client (used by the Python etcd library) doesn't allow untrusted
    # etcd certificates, although traefik does allow them.
    etcd_insecure_skip_verify = Bool(
        False,
        config=True,
        help="""Traefik will by default validate SSL certificate of etcd backend""",
    )

    grpc_options = List(
        config=True,
        allow_none=True,
        default_value=None,
        help="""Any grpc options that need to be passed to the etcd client""",
    )

    @default("executor")
    def _default_executor(self):
        return ThreadPoolExecutor(1)

    etcd_url = Unicode(
        "http://127.0.0.1:2379",
        config=True,
        help="URL for the etcd endpoint.",
    )

    etcd_username = Unicode(
        "",
        config=True,
        help="Username for accessing etcd.",
    )
    etcd_password = Unicode(
        "",
        config=True,
        help="Password for accessing etcd.",
    )

    kv_url = Unicode("DEPRECATED", config=True).tag(
        deprecated_in="1.0",
        deprecated_for="etcd_url",
    )
    kv_username = Unicode("DEPRECATED", config=True).tag(
        deprecated_in="1.0",
        deprecated_for="etcd_username",
    )
    kv_password = Unicode("DEPRECATED", config=True).tag(
        deprecated_in="1.0",
        deprecated_for="etcd_password",
    )

    etcd = Any()

    @default("etcd")
    def _default_client(self):
        etcd_service = urlparse(self.etcd_url)
        try:
            import etcd3
        except ImportError:
            raise ImportError(
                "Please install etcd3 or etcdpy package to use traefik-proxy with etcd3"
            )
        kwargs = {
            'host': etcd_service.hostname,
            'port': etcd_service.port,
            'ca_cert': self.etcd_client_ca_cert,
            'cert_cert': self.etcd_client_cert_crt,
            'cert_key': self.etcd_client_cert_key,
            'grpc_options': self.grpc_options,
        }
        if self.etcd_password:
            kwargs.update(
                {
                    "user": self.etcd_username,
                    "password": self.etcd_password,
                }
            )
        return etcd3.client(**kwargs)

    def _cleanup(self):
        super()._cleanup()
        self.etcd.close()

    # low-level etcd APIs
    @run_on_executor
    def _etcd_transaction(self, success_actions):
        status, response = self.etcd.transaction(
            compare=[], success=success_actions, failure=[]
        )
        if status != True:
            raise RuntimeError(f"etcd transaction failed: {status}: {response}")
        return response

    @run_on_executor
    def _etcd_get(self, key):
        value, _ = self.etcd.get(key)
        return value

    @run_on_executor
    def _etcd_get_prefix(self, prefix):
        if not prefix.endswith(self.kv_separator):
            prefix += self.kv_separator
        data = list(self.etcd.get_prefix(prefix))
        return data

    # key-value generic methods

    async def _kv_get_tree(self, prefix):
        keys_values = [
            (meta.key.decode("utf8"), value.decode("utf8"))
            for value, meta in await self._etcd_get_prefix(prefix)
        ]
        return self.unflatten_dict_from_kv(keys_values, root_key=prefix)

    async def _kv_atomic_set(self, to_set):
        transactions = []
        for k, v in to_set.items():
            transactions.append(self.etcd.transactions.put(k, v))
        await self._etcd_transaction(transactions)

    async def _kv_atomic_delete(self, *keys):
        """Delete one or more keys from the kv store"""

        transactions = []
        delete = self.etcd.transactions.delete

        for key in keys:
            if key.endswith(self.kv_separator):
                # it's a tree, we have to list sub-keys to delete them atomically
                for _, meta in await self._etcd_get_prefix(key):
                    transactions.append(delete(meta.key))
            else:
                transactions.append(delete(key))
        await self._etcd_transaction(transactions)

    # traefik + etcd methods
    def _setup_traefik_static_config(self):
        self.log.debug("Setting up the etcd provider in the static config")
        url = urlparse(self.etcd_url)
        self.static_config.update(
            {
                "providers": {
                    "etcd": {
                        "endpoints": [url.netloc],
                        "rootKey": self.kv_traefik_prefix,
                    }
                }
            }
        )
        if url.scheme == "https":
            # If etcd is running over TLS, then traefik needs to know
            tls_conf = {}
            if self.etcd_client_ca_cert is not None:
                tls_conf["ca"] = self.etcd_client_ca_cert
            tls_conf["insecureSkipVerify"] = self.etcd_insecure_skip_verify
            self.static_config["providers"]["etcd"]["tls"] = tls_conf

        if self.etcd_username and self.etcd_password:
            self.static_config["providers"]["etcd"].update(
                {
                    "username": self.etcd_username,
                    "password": self.etcd_password,
                }
            )
        return super()._setup_traefik_static_config()



"""Tests for the treafik proxy implementations"""

import asyncio
import copy
import inspect
import pprint
import ssl
import subprocess
import sys
from contextlib import contextmanager
from os.path import abspath, dirname, join
from random import randint
from unittest.mock import Mock
from urllib.parse import quote, urlparse

import pytest
import utils
import websockets
from jupyterhub.objects import Hub, Server
from jupyterhub.user import User
from jupyterhub.utils import exponential_backoff, url_path_join
from tornado.httpclient import AsyncHTTPClient, HTTPClientError, HTTPRequest

from jupyterhub_traefik_proxy.proxy import TraefikProxy

# Mark all tests in this file as slow
pytestmark = [pytest.mark.slow]

pp = pprint.PrettyPrinter(indent=2)


class MockApp:
    def __init__(self):
        self.hub = Hub(routespec="/")


class MockSpawner:
    name = ""
    server = None
    pending = None

    def __init__(self, name="", *, user, **kwargs):
        self.name = name
        self.user = user
        for key, value in kwargs.items():
            setattr(self, key, value)
        self.proxy_spec = url_path_join(self.user.proxy_spec, name, "/")

    def start(self):
        self.server = Server.from_url("http://127.0.0.1:%i" % randint(1025, 65535))

    def stop(self):
        self.server = None

    @property
    def ready(self):
        """Is this server ready to use?

        A server is not ready if an event is pending.
        """
        return self.server is not None

    @property
    def active(self):
        """Return True if the server is active.

        This includes fully running and ready or any pending start/stop event.
        """
        return self.ready


class MockUser(User):
    """Mock User for use in proxytest"""

    def __init__(self, name):
        orm_user = Mock()
        orm_user.name = name
        orm_user.orm_spawners = ""
        super().__init__(orm_user=orm_user, db=Mock())

    def _new_spawner(self, spawner_name, **kwargs):
        return MockSpawner(spawner_name, user=self, **kwargs)


def assert_equal(value, expected):
    try:
        assert value == expected
    except AssertionError:
        pp.pprint({'value': value})
        pp.pprint({"expected": expected})
        raise


@pytest.fixture(scope="session")
def launch_backends():
    """Launch n backends

    Session-scoped, so backends are re-used

    The fixture result is an async function that takes a number of backends required,
    and returns a list of URLs for those backends.

    When the function returns, the backends are already running and responsive.
    """

    dummy_server_path = abspath(join(dirname(__file__), "dummy_http_server.py"))
    running_backends = []
    urls = []
    base_port = 9000

    async def _launch_backends(n=1):
        """Launch `n` backends, returning their URLs

        Always returns a list of length `n`.
        """
        already_available = len(running_backends)
        for i in range(already_available, n):
            port = base_port + i
            url = f"http://127.0.0.1:{port}"
            print(f"Launching backend on {url}")
            backend = subprocess.Popen([sys.executable, dummy_server_path, str(port)])
            running_backends.append(backend)
            urls.append(url)

        if already_available < n:
            # await _new_ backends
            await wait_for_services(urls[already_available:])

        return urls[:n]

    yield _launch_backends

    for proc in running_backends:
        proc.terminate()
    for proc in running_backends:
        proc.communicate()
    for proc in running_backends:
        proc.wait()


async def wait_for_services(urls):
    # Wait until traefik and the backend are ready
    await exponential_backoff(
        utils.check_services_ready, "Service not reacheable", urls=urls
    )


def test_default_port():
    p = TraefikProxy(
        public_url="http://127.0.0.1/", traefik_api_url="https://127.0.0.1/"
    )
    assert p.public_url == "http://127.0.0.1:80/"
    assert p.traefik_api_url == "https://127.0.0.1:443/"

    with pytest.raises(ValueError):
        TraefikProxy(public_url="ftp://127.0.0.1:23/")


@pytest.mark.parametrize(
    "routespec, existing_routes",
    [
        # default route
        (
            "/",
            [
                "/abc",
                "/has%20space/",
                "/has%20space/foo/",
                "/missing-trailing/",
                "/missing-trailing/slash",
                "/has/",
                "/has/@/",
                "host.name/",
                "host.name/path/",
                "other.host/",
                "other.host/path/",
                "other.host/path/no/",
                "other.host/path/no/slash",
            ],
        ),
        ("/has%20space/foo/", ["/", "/has%20space/", "/has%20space/foo/abc/"]),
        (
            "/missing-trailing/slash",
            ["/", "/missing-trailing/", "/missing-trailing/slash/abc"],
        ),
        ("/has/@/", ["/", "/has/", "/has/@/abc/"]),
        (
            "/has/" + quote("üñîçø∂é"),
            ["/", "/has/", "/has/" + quote("üñîçø∂é") + "/abc/"],
        ),
        ("host.name/path/", ["/", "host.name/", "host.name/path/abc/"]),
        (
            "other.host/path/no/slash",
            [
                "/",
                "other.host/",
                "other.host/path/",
                "other.host/path/no/",
                "other.host/path/no/slash/abc/",
            ],
        ),
        (
            "/one/",
            [],
        ),
    ],
)
async def test_add_get_delete(
    request, proxy, launch_backends, routespec, existing_routes, event_loop
):
    data = {"test": "test1", "user": "username"}

    backends = await launch_backends(1 + len(existing_routes))

    default_backend = backends[0]
    extra_backends = backends[1:]

    proxy_url = proxy.public_url.rstrip("/")

    def normalize_spec(spec):
        return proxy.validate_routespec(spec)

    def expected_output(spec, url):
        return {
            "routespec": normalize_spec(spec),
            "target": url,
            "data": data,
        }

    # just use existing Jupyterhub check instead of making own one
    def expect_value_error(spec):
        try:
            normalize_spec(spec)
        except ValueError:
            return True

        return False

    @contextmanager
    def context(spec):
        if expect_value_error(spec):
            with pytest.raises(ValueError):
                yield
        else:
            yield

    async def test_route_exist(spec, backend):
        with context(spec):
            route = await proxy.get_route(spec)

        if not expect_value_error(spec):
            try:
                del route["data"]["last_activity"]  # CHP
            except TypeError as e:
                raise TypeError(f"{e}\nRoute got:{route}")
            except KeyError:
                pass

            assert_equal(route, expected_output(spec, backend))

            # Test the actual routing
            responding_backend1 = await utils.get_responding_backend_port(
                proxy_url, normalize_spec(spec)
            )
            responding_backend2 = await utils.get_responding_backend_port(
                proxy_url, normalize_spec(spec) + "something"
            )
            assert responding_backend1 == urlparse(backend).port
            assert responding_backend2 == urlparse(backend).port

    # Create existing routes
    futures = []
    for i, spec in enumerate(existing_routes):
        f = proxy.add_route(spec, extra_backends[i], copy.copy(data))
        futures.append(f)

    if futures:
        await asyncio.gather(*futures)

    def finalizer():
        async def cleanup():
            """Cleanup"""
            futures = []
            for spec in existing_routes:
                try:
                    f = proxy.delete_route(spec)
                    futures.append(f)
                except Exception:
                    pass
            if futures:
                await asyncio.gather(*futures)

        event_loop.run_until_complete(cleanup())

    request.addfinalizer(finalizer)

    # Test add
    with context(routespec):
        await proxy.add_route(routespec, default_backend, copy.copy(data))

    # Test get
    await test_route_exist(routespec, default_backend)
    for i, spec in enumerate(existing_routes):
        await test_route_exist(spec, extra_backends[i])

    # Test delete
    with context(routespec):
        await proxy.delete_route(routespec)
        route = await proxy.get_route(routespec)

    # Test that deleted route does not exist anymore
    if not expect_value_error(routespec):
        assert_equal(route, None)

        async def _wait_for_deletion():
            deleted = 0
            for spec in [
                normalize_spec(routespec),
                normalize_spec(routespec) + "something",
            ]:
                try:
                    result = await utils.get_responding_backend_port(proxy_url, spec)
                    if result != urlparse(default_backend).port:
                        deleted += 1
                except HTTPClientError:
                    deleted += 1

            return deleted == 2

        # If this raises a TimeoutError, the route wasn't properly deleted,
        # thus the proxy still has a route for the given routespec
        await exponential_backoff(_wait_for_deletion, "Route still exists")

    # Test that other routes are still exist
    for i, spec in enumerate(existing_routes):
        await test_route_exist(spec, extra_backends[i])


async def test_get_all_routes(proxy, launch_backends):
    # initial state: no routes
    routes = await proxy.get_all_routes()
    assert routes == {}

    routespecs = ["/proxy/path1", "/proxy/path2/", "/proxy/path3/"]
    targets = await launch_backends(len(routespecs))
    datas = [{"test": "test1"}, {}, {"test": "test2"}]

    expected_output = {
        routespecs[0]
        + "/": {
            "routespec": routespecs[0] + "/",
            "target": targets[0],
            "data": datas[0],
        },
        routespecs[1]: {
            "routespec": routespecs[1],
            "target": targets[1],
            "data": datas[1],
        },
        routespecs[2]: {
            "routespec": routespecs[2],
            "target": targets[2],
            "data": datas[2],
        },
    }

    futures = []
    for routespec, target, data in zip(routespecs, targets, datas):
        f = proxy.add_route(routespec, target, copy.copy(data))
        futures.append(f)
    if futures:
        await asyncio.gather(*futures)

    routes = await proxy.get_all_routes()
    try:
        for route_key in routes.keys():
            del routes[route_key]["data"]["last_activity"]  # CHP
    except KeyError:
        pass

    assert_equal(routes, expected_output)

    for routespec in routespecs:
        await proxy.delete_route(routespec)
    routes = await proxy.get_all_routes()
    assert routes == {}


async def test_host_origin_headers(proxy, launch_backends):
    routespec = "/user/username/"
    target = "http://127.0.0.1:9000"
    data = {}

    proxy_url = urlparse(proxy.public_url)
    traefik_port = proxy_url.port
    traefik_host = proxy_url.hostname
    backends = await launch_backends(1)
    target = backends[0]
    urlparse(target).port

    # wait for traefik to be reachable
    await exponential_backoff(
        utils.check_host_up_http,
        "Traefik not reacheable",
        url=proxy_url.geturl(),
    )

    # wait for backend to be reachable
    await exponential_backoff(
        utils.check_host_up_http,
        "Backends not reacheable",
        url=target,
    )

    # Add route to default_backend
    await proxy.add_route(routespec, target, data)

    req_url = proxy.public_url.rstrip("/") + routespec

    expected_host_header = traefik_host + ":" + str(traefik_port)
    expected_origin_header = proxy.public_url + routespec

    req = HTTPRequest(
        req_url,
        method="GET",
        headers={"Host": expected_host_header, "Origin": expected_origin_header},
        validate_cert=False,
    )
    resp = await AsyncHTTPClient().fetch(req)

    assert resp.headers["Host"] == expected_host_header
    assert resp.headers["Origin"] == expected_origin_header


@pytest.mark.parametrize("username", ["zoe", "50fia", "秀樹", "~TestJH", "has@"])
async def test_check_routes(proxy, username):
    # fill out necessary attributes for check_routes
    proxy.app = MockApp()
    proxy.hub = proxy.app.hub

    users = {}
    services = {}
    # run initial check first, to ensure that `/` is in the routes
    await proxy.check_routes(users, services)
    routes = await proxy.get_all_routes()
    assert sorted(routes) == ["/"]

    users[username] = test_user = MockUser(username)
    spawner = test_user.spawners[""]
    f = spawner.start()
    if inspect.isawaitable(f):
        await f
    assert spawner.ready
    assert spawner.active
    await proxy.add_user(test_user, "")

    # check a valid route exists for user
    routes = await proxy.get_all_routes()
    before = sorted(routes)
    assert test_user.proxy_spec in before

    # check if a route is removed when user deleted
    await proxy.check_routes(users, services)
    await proxy.delete_user(test_user)
    routes = await proxy.get_all_routes()
    during = sorted(routes)
    assert test_user.proxy_spec not in during

    # check if a route exists for user
    await proxy.check_routes(users, services)
    routes = await proxy.get_all_routes()
    after = sorted(routes)
    assert test_user.proxy_spec in after

    # check that before and after state are the same
    assert_equal(before, after)


async def test_websockets(proxy, launch_backends):
    routespec = "/user/username/"
    data = {}

    proxy_url = urlparse(proxy.public_url)
    proxy_url.port
    proxy_url.hostname
    backends = await launch_backends(1)
    target = backends[0]
    default_backend_port = urlparse(target).port

    # Add route to default_backend
    await proxy.add_route(routespec, target, data)

    if proxy.is_https:
        kwargs = {'ssl': ssl._create_unverified_context()}
        scheme = "wss://"
    else:
        kwargs = {}
        scheme = "ws://"
    req_url = scheme + proxy_url.netloc + url_path_join(routespec, "ws")

    # Don't validate the ssl certificate, it's self-signed by traefik
    print(f"Connecting with websockets to {req_url}")
    async with websockets.connect(req_url, **kwargs) as websocket:
        port = await websocket.recv()

    assert port == str(default_backend_port)



import pytest

from jupyterhub_traefik_proxy.kv_proxy import TKvProxy


@pytest.mark.parametrize(
    "orig, expected",
    [
        ({}, {}),
        (
            {"key": "value"},
            {"key": "value"},
        ),
        (
            {"key": 1},
            {"key": "1"},
        ),
        (
            {"key": 1.5},
            {"key": "1.5"},
        ),
        (
            {"key": ["a", "b"]},
            {"key/0": "a", "key/1": "b"},
        ),
        (
            {"key": {"level": True, "deeper": {"anddeeper": False}}},
            {"key/level": "true", "key/deeper/anddeeper": "false"},
        ),
    ],
)
def test_flatten_dict(orig, expected):
    proxy = TKvProxy()
    flat = proxy.flatten_dict_for_kv(orig)
    assert flat == expected


@pytest.mark.parametrize(
    "orig, expected",
    [
        ({"key": object()}, ValueError),
    ],
)
def test_flatten_dict_error(orig, expected):
    proxy = TKvProxy()
    with pytest.raises(expected):
        proxy.flatten_dict_for_kv(orig)


@pytest.mark.parametrize(
    "flat, root_key, expected",
    [
        ([], "", {}),
        (
            [("key", "value")],
            "",
            {"key": "value"},
        ),
        (
            [("key", "value")],
            "nosuchkey",
            {},
        ),
        (
            [("key", "1")],
            "",
            {"key": "1"},
        ),
        (
            [("key/0", "a"), ("key/1", "b")],
            "",
            {"key": ["a", "b"]},
        ),
        (
            [("key/level", "true"), ("key/deeper/anddeeper", "false")],
            "",
            {"key": {"level": "true", "deeper": {"anddeeper": "false"}}},
        ),
        (
            [("key/level", "true"), ("key/deeper/anddeeper", "false")],
            "key/deeper",
            {"anddeeper": "false"},
        ),
    ],
)
def test_unflatten_dict(flat, root_key, expected):
    proxy = TKvProxy()
    unflat = proxy.unflatten_dict_from_kv(flat, root_key=root_key)
    assert unflat == expected


@pytest.mark.parametrize(
    "flat, expected",
    [
        (
            [("key/1", "value")],
            IndexError,
        ),
    ],
)
def test_unflatten_dict_error(flat, expected):
    proxy = TKvProxy()
    with pytest.raises(expected):
        proxy.unflatten_dict_from_kv(flat)



"""Traefik implementation

Custom proxy implementations can subclass :class:`Proxy`
and register in JupyterHub config:

.. sourcecode:: python

    from mymodule import MyProxy
    c.JupyterHub.proxy_class = MyProxy

Route Specification:

- A routespec is a URL prefix ([host]/path/), e.g.
  'host.tld/path/' for host-based routing or '/path/' for default routing.
- Route paths should be normalized to always start and end with '/'
"""

# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

import asyncio
import json
import os
import ssl
from os.path import abspath
from subprocess import Popen, TimeoutExpired
from urllib.parse import urlparse, urlunparse

from jupyterhub.proxy import Proxy
from jupyterhub.utils import exponential_backoff, new_token, url_path_join
from tornado.httpclient import AsyncHTTPClient, HTTPClientError
from traitlets import Any, Bool, Dict, Integer, Unicode, default, observe, validate

from . import traefik_utils


class TraefikProxy(Proxy):
    """JupyterHub Proxy implementation using traefik"""

    provider_name = ""  # must be set in subclasses

    traefik_process = Any()

    concurrency = Integer(
        10,
        config=True,
        help="""
        The number of requests allowed to be concurrently outstanding to the proxy

        Limiting this number avoids potential timeout errors
        by sending too many requests to update the proxy at once
        """,
    )
    semaphore = Any()

    @default('semaphore')
    def _default_semaphore(self):
        return asyncio.BoundedSemaphore(self.concurrency)

    @observe('concurrency')
    def _concurrency_changed(self, change):
        self.semaphore = asyncio.BoundedSemaphore(change.new)

    static_config_file = Unicode(
        "traefik.toml", config=True, help="""traefik's static configuration file"""
    )

    toml_static_config_file = Unicode(
        config=True,
        help="Deprecated. Use static_config_file",
    ).tag(
        deprecated_in="1.0",
        deprecated_for="static_config_file",
    )

    def _deprecated_trait(self, change):
        """observer for deprecated traits"""
        trait = change.owner.traits()[change.name]
        old_attr = change.name
        new_attr = trait.metadata["deprecated_for"]
        version = trait.metadata["deprecated_in"]
        if "." in new_attr:
            new_cls_attr = new_attr
            new_attr = new_attr.rsplit(".", 1)[1]
        else:
            new_cls_attr = f"{self.__class__.__name__}.{new_attr}"

        new_value = getattr(self, new_attr)
        if new_value != change.new:
            # only warn if different
            # protects backward-compatible config from warnings
            # if they set the same value under both names
            message = "{cls}.{old} is deprecated in {cls} {version}, use {new} instead".format(
                cls=self.__class__.__name__,
                old=old_attr,
                new=new_cls_attr,
                version=version,
            )
            self.log.warning(message)

            setattr(self, new_attr, change.new)

    def __init__(self, **kwargs):
        # observe deprecated config names in oauthenticator
        for name, trait in self.class_traits().items():
            if trait.metadata.get("deprecated_in"):
                self.observe(self._deprecated_trait, name)
        super().__init__(**kwargs)
        # not calling .start, make sure we load our initial dynamic config
        # and that our static config matches what we expect
        # can't await here because we're not async,
        # so await in our Proxy methods
        if not self.should_start:
            self._start_future = asyncio.ensure_future(self._start_external())
        else:
            self._start_future = None

    static_config = Dict()
    dynamic_config = Dict()

    traefik_providers_throttle_duration = Unicode(
        "0s",
        config=True,
        help="""
            throttle traefik reloads of configuration.

            When traefik sees a change in configuration,
            it will wait this long before applying the next one.
            This affects how long adding a user to the proxy will take.

            See https://doc.traefik.io/traefik/providers/overview/#providersprovidersthrottleduration
            """,
    )

    traefik_api_url = Unicode(
        "http://localhost:8099",
        config=True,
        help="""traefik authenticated api endpoint url""",
    )

    traefik_api_validate_cert = Bool(
        True,
        config=True,
        help="""validate SSL certificate of traefik api endpoint""",
    )

    traefik_log_level = Unicode(config=True, help="""traefik's log level""")

    traefik_api_password = Unicode(
        config=True, help="""The password for traefik api login"""
    )

    traefik_env = Dict(
        config=True,
        help="""Environment variables to set for the traefik process.

        Only has an effect when traefik is a subprocess (should_start=True).
        """,
    )

    provider_name = Unicode(
        help="""The provider name that Traefik expects, e.g. file, consul, etcd"""
    )

    is_https = Bool(
        help="""Whether :attr:`.public_url` specifies an https entrypoint"""
    )

    @default("is_https")
    def get_is_https(self):
        # Check if we set https
        return urlparse(self.public_url).scheme == "https"

    @validate("public_url", "traefik_api_url")
    def _add_port(self, proposal):
        url = proposal.value
        parsed = urlparse(url)
        if parsed.scheme not in ("http", "https"):
            raise ValueError(
                f"{self.__class__.__name__}.{proposal.trait.name} must be of the form http[s]://host:port/, got {url}"
            )
        if not parsed.port:
            # ensure port is defined
            if parsed.scheme == 'http':
                parsed = parsed._replace(netloc=f'{parsed.hostname}:80')
            elif parsed.scheme == 'https':
                parsed = parsed._replace(netloc=f'{parsed.hostname}:443')
            url = urlunparse(parsed)
        return url

    traefik_cert_resolver = Unicode(
        config=True,
        help="""The traefik certificate Resolver to use for requesting certificates""",
    )

    # FIXME: How best to enable TLS on routers assigned to only select
    # entrypoints defined here?
    traefik_entrypoint = Unicode(
        help="""The traefik entrypoint name to use.

        By default, will be `web` if http or `websecure` if https.

        If running traefik externally with your own specified entrypoint name,
        set this value.
        """,
        config=True,
    )

    @default("traefik_entrypoint")
    def _default_traefik_entrypoint(self):
        """Find the traefik entrypoint that matches our :attrib:`self.public_url`"""
        if self.is_https:
            return "websecure"
        else:
            return "web"

    @default("traefik_api_password")
    def _warn_empty_password(self):
        self.log.warning("Traefik API password was not set.")

        if self.should_start:
            # Generating tokens is fine if the Hub is starting the proxy
            self.log.warning("Generating a random token for traefik_api_username...")
            return new_token()

        self.log.warning(
            "Please set c.TraefikProxy.traefik_api_password to authenticate with traefik"
            " if the proxy was not started by the Hub."
        )
        return ""

    traefik_api_username = Unicode(
        config=True, help="""The username for traefik api login"""
    )

    @default("traefik_api_username")
    def _warn_empty_username(self):
        self.log.warning("Traefik API username was not set.")

        if self.should_start:
            self.log.warning('Defaulting traefik_api_username to "jupyterhub"')
            return "jupyterhub"

        self.log.warning(
            "Please set c.TraefikProxy.traefik_api_username to authenticate with traefik"
            " if the proxy was not started by the Hub."
        )
        return ""

    traefik_api_hashed_password = Unicode()

    check_route_timeout = Integer(
        30,
        config=True,
        help="""Timeout (in seconds) when waiting for traefik to register an updated route.""",
    )

    def _generate_htpassword(self):
        from passlib.hash import apr_md5_crypt

        self.traefik_api_hashed_password = apr_md5_crypt.hash(self.traefik_api_password)

    async def _check_for_traefik_service(self, routespec, kind):
        """Check for an expected router or service in the Traefik API.

        This is used to wait for traefik to load configuration
        from a provider
        """
        # expected e.g. 'service' + '_' + routespec @ file
        routespec = self.validate_routespec(routespec)
        expected = (
            traefik_utils.generate_alias(routespec, kind) + "@" + self.provider_name
        )
        path = f"/api/http/{kind}s/{expected}"
        try:
            resp = await self._traefik_api_request(path)
            json.loads(resp.body)
        except HTTPClientError as e:
            if e.code == 404:
                self.log.debug(f"traefik {expected} not yet in {kind}")
                return False
            self.log.exception(f"Error checking traefik api for {kind} {routespec}")
            return False
        except Exception:
            self.log.exception(f"Error checking traefik api for {kind} {routespec}")
            return False

        # found the expected endpoint
        return True

    async def _wait_for_route(self, routespec):
        self.log.debug("Waiting for %s to register with traefik", routespec)

        async def _check_traefik_dynamic_conf_ready():
            """Check if traefik loaded its dynamic configuration yet"""
            if not await self._check_for_traefik_service(routespec, "service"):
                return False
            if not await self._check_for_traefik_service(routespec, "router"):
                return False

            return True

        await exponential_backoff(
            _check_traefik_dynamic_conf_ready,
            f"Traefik route for {routespec} configuration not available",
            scale_factor=1.2,
            timeout=self.check_route_timeout,
        )

    async def _traefik_api_request(self, path):
        """Make an API request to traefik"""
        url = url_path_join(self.traefik_api_url, path)
        self.log.debug("Fetching traefik api %s", url)
        resp = await AsyncHTTPClient().fetch(
            url,
            auth_username=self.traefik_api_username,
            auth_password=self.traefik_api_password,
            validate_cert=self.traefik_api_validate_cert,
        )
        if resp.code >= 300:
            self.log.warning("%s GET %s", resp.code, url)
        else:
            self.log.debug("%s GET %s", resp.code, url)
        return resp

    async def _wait_for_static_config(self):
        async def _check_traefik_static_conf_ready():
            """Check if traefik loaded its static configuration yet"""
            try:
                await self._traefik_api_request("/api/overview")
                await self._traefik_api_request(
                    f"/api/entrypoints/{self.traefik_entrypoint}"
                )
            except ConnectionRefusedError:
                self.log.debug(
                    f"Connection Refused waiting for traefik at {self.traefik_api_url}. It's probably starting up..."
                )
                return False
            except HTTPClientError as e:
                if e.code == 599:
                    self.log.debug(
                        f"Connection error waiting for traefik at {self.traefik_api_url}. It's probably starting up..."
                    )
                    return False
                if e.code == 404:
                    if "/entrypoints/" in e.response.request.url:
                        self.log.warning(
                            f"c.{self.__class__.__name__}.traefik_entrypoint={self.traefik_entrypoint!r} not found in traefik. Is it correct?"
                        )
                    else:
                        self.log.debug(
                            f"traefik api at {e.response.request.url} overview not ready yet"
                        )
                    return False
                # unexpected
                self.log.error(f"Error checking for traefik static configuration {e}")
                return False
            except (OSError, ssl.SSLError) as e:
                # Can occur if SSL isn't set up yet
                self.log.warning(
                    f"SSL Error checking for traefik static configuration: {e}"
                )
                return False

            return True

        await exponential_backoff(
            _check_traefik_static_conf_ready,
            "Traefik static configuration not available",
            timeout=self.check_route_timeout,
        )

    def _stop_traefik(self):
        self.log.info("Cleaning up traefik proxy [pid=%i]...", self.traefik_process.pid)
        self.traefik_process.terminate()
        try:
            self.traefik_process.communicate(timeout=10)
        except TimeoutExpired:
            self.traefik_process.kill()
            self.traefik_process.communicate()
        finally:
            self.traefik_process.wait()

    def _start_traefik(self):
        env = os.environ.copy()
        env.update(self.traefik_env)
        try:
            self.traefik_process = Popen(
                ["traefik", "--configfile", abspath(self.static_config_file)],
                env=env,
            )
        except FileNotFoundError:
            self.log.error(
                "Failed to find traefik\n"
                "The proxy can be downloaded from https://github.com/traefik/traefik/releases/."
            )
            raise

    async def _setup_traefik_static_config(self):
        """When should_start=True, we are in control of traefik's static configuration
        file. This sets up the entrypoints and api handler in self.static_config, and
        then saves it to :attrib:`self.static_config_file`.

        Subclasses should specify any traefik providers themselves, in
        :attrib:`self.static_config["providers"]`
        """
        self.static_config["providers"][
            "providersThrottleDuration"
        ] = self.traefik_providers_throttle_duration

        if self.traefik_log_level:
            self.static_config["log"] = {"level": self.traefik_log_level}

        entrypoints = {
            self.traefik_entrypoint: {
                "address": urlparse(self.public_url).netloc,
            },
            "enter_api": {
                "address": urlparse(self.traefik_api_url).netloc,
            },
        }

        self.static_config["entryPoints"] = entrypoints
        self.static_config["api"] = {}

        self.log.info(f"Writing traefik static config: {self.static_config}")

        try:
            handler = traefik_utils.TraefikConfigFileHandler(self.static_config_file)
            handler.atomic_dump(self.static_config)
        except Exception:
            self.log.error("Couldn't set up traefik's static config.")
            raise

    async def _setup_traefik_dynamic_config(self):
        self.log.debug("Setting up traefik's dynamic config...")
        self._generate_htpassword()
        api_url = urlparse(self.traefik_api_url)
        api_path = api_url.path if api_url.path else '/api'
        api_credentials = (
            f"{self.traefik_api_username}:{self.traefik_api_hashed_password}"
        )
        http = self.dynamic_config.setdefault("http", {})
        routers = http.setdefault("routers", {})
        routers["route_api"] = {
            "rule": f"Host(`{api_url.hostname}`) && PathPrefix(`{api_path}`)",
            "entryPoints": ["enter_api"],
            "service": "api@internal",
            "middlewares": ["auth_api"],
        }
        middlewares = http.setdefault("middlewares", {})
        middlewares["auth_api"] = {"basicAuth": {"users": [api_credentials]}}
        if self.ssl_cert and self.ssl_key:
            tls = self.dynamic_config.setdefault("tls", {})
            stores = tls.setdefault("stores", {})
            stores["default"] = {
                "defaultCertificate": {
                    "certFile": self.ssl_cert,
                    "keyFile": self.ssl_key,
                }
            }
        await self._apply_dynamic_config(self.dynamic_config, None)

    def validate_routespec(self, routespec):
        """Override jupyterhub's default Proxy.validate_routespec method, as traefik
        can set router rule's on both Host and PathPrefix rules combined.
        """
        if not routespec.endswith("/"):
            routespec = routespec + "/"
        return routespec

    async def start(self):
        """Start the proxy.

        Will be called during startup if should_start is True.

        **Subclasses must define this method**
        if the proxy is to be started by the Hub
        """
        await self._setup_traefik_static_config()
        await self._setup_traefik_dynamic_config()
        self._start_traefik()
        await self._wait_for_static_config()

    async def _start_external(self):
        """Startup function called when `not self.should_start`

        Ensures dynamic config is setup and static config is loaded
        """
        await self._setup_traefik_dynamic_config()
        await self._wait_for_static_config()
        self._start_future = None

    async def stop(self):
        """Stop the proxy.

        Will be called during teardown if should_start is True.

        **Subclasses must define this method**
        if the proxy is to be started by the Hub
        """
        self._stop_traefik()
        self._cleanup()

    def _cleanup(self):
        """Cleanup after stop

        Extend if there's more to cleanup than the static config file
        """
        if self.should_start:
            try:
                os.remove(self.static_config_file)
            except Exception as e:
                self.log.error(
                    f"Failed to remove traefik config file {self.static_config_file}: {e}"
                )

    def _dynamic_config_for_route(self, routespec, target, data):
        """Returns two dicts, which will be used to update traefik configuration for a given route

        (traefik_config, jupyterhub_config) -
            where traefik_config is traefik dynamic_config to be merged,
            and jupyterhub_config is jupyterhub-specific data to be stored elsewhere
            (implementation-specific) and associated with the route
        """

        service_alias = traefik_utils.generate_alias(routespec, "service")
        router_alias = traefik_utils.generate_alias(routespec, "router")
        rule = traefik_utils.generate_rule(routespec)
        # dynamic config to deep merge
        traefik_config = {
            "http": {
                "routers": {},
                "services": {},
            },
        }
        jupyterhub_config = {
            "routes": {},
        }
        traefik_config["http"]["routers"][router_alias] = router = {
            "service": service_alias,
            "rule": rule,
            "entryPoints": [self.traefik_entrypoint],
        }
        traefik_config["http"]["services"][service_alias] = {
            "loadBalancer": {"servers": [{"url": target}], "passHostHeader": True}
        }

        # Enable TLS on this router if globally enabled
        if self.is_https:
            tls_config = {}
            if self.traefik_cert_resolver:
                tls_config["certResolver"] = self.traefik_cert_resolver
            else:
                # we need _some_ key to be set
                # because key-value stores can't store empty dicts.
                # put a default value here
                tls_config["options"] = "default"

            router["tls"] = tls_config

        # Add the data node to a separate top-level node, so traefik doesn't see it.
        # key needs to be key-value safe (no '/')
        # store original routespec, router, service aliases for easy lookup
        jupyterhub_config["routes"][router_alias] = {
            "data": data,
            "routespec": routespec,
            "target": target,
            "router": router_alias,
            "service": service_alias,
        }
        return traefik_config, jupyterhub_config

    async def add_route(self, routespec, target, data):
        """Add a route to the proxy.

        **Subclasses must define this method**

        Args:
            routespec (str): A URL prefix ([host]/path/) for which this route will be matched,
                e.g. host.name/path/
            target (str): A full URL that will be the target of this route.
            data (dict): A JSONable dict that will be associated with this route, and will
                be returned when retrieving information about this route.

        Will raise an appropriate Exception (FIXME: find what?) if the route could
        not be added.

        The proxy implementation should also have a way to associate the fact that a
        route came from JupyterHub.
        """
        if self._start_future and not self._start_future.done():
            await self._start_future
        routespec = self.validate_routespec(routespec)

        traefik_config, jupyterhub_config = self._dynamic_config_for_route(
            routespec, target, data
        )

        try:
            async with self.semaphore:
                await self._apply_dynamic_config(traefik_config, jupyterhub_config)
                await self._wait_for_route(routespec)
        except TimeoutError:
            self.log.error(f"Traefik route for {routespec} never appeared.")
            raise

    def _keys_for_route(self, routespec):
        """Return (traefik_keys, jupyterhub_keys)

        keys in dynamic_config and jupyterhub_dynamic_config that correspond to a route.

        These keys may be used in delete_route and get_route.

        each key is a list of strings, representing
        `config[key[0]][key[1]]` etc.

        i.e. ( (["http", "routers", "router_name"], ("routes", "route_name") )
        """
        service_alias = traefik_utils.generate_alias(routespec, "service")
        router_alias = traefik_utils.generate_alias(routespec, "router")
        traefik_keys = (
            ["http", "routers", router_alias],
            ["http", "services", service_alias],
        )
        jupyterhub_keys = (["routes", router_alias],)
        return traefik_keys, jupyterhub_keys

    async def _delete_dynamic_config(self, traefik_keys, jupyterhub_keys):
        """Delete keys from dynamic configuration

        Must be implemented in subclasses
        """
        raise NotImplementedError()

    async def delete_route(self, routespec):
        """Delete a route with a given routespec if it exists."""
        routespec = self.validate_routespec(routespec)
        traefik_keys, jupyterhub_keys = self._keys_for_route(routespec)
        await self._delete_dynamic_config(traefik_keys, jupyterhub_keys)
        self.log.debug("Route %s was deleted.", routespec)

    async def _get_jupyterhub_dynamic_config(self):
        """Get the jupyterhub part of our dynamic config

        This houses serialization of routespecs, etc.
        """
        raise NotImplementedError()

    async def check_routes(self, *args, **kwargs):
        if self._start_future and not self._start_future.done():
            await self._start_future
        return await super().check_routes(*args, **kwargs)

    async def get_all_routes(self):
        """Fetch and return all the routes associated by JupyterHub from the
        proxy.

        **Subclasses must define this method**

        Should return a dictionary of routes, where the keys are
        routespecs and each value is a dict of the form::

          {
            'routespec': the route specification ([host]/path/)
            'target': the target host URL (proto://host) for this route
            'data': the attached data dict for this route (as specified in add_route)
          }
        """
        if self._start_future and not self._start_future.done():
            await self._start_future

        jupyterhub_config = await self._get_jupyterhub_dynamic_config()

        all_routes = {}
        for _key, route in jupyterhub_config.get("routes", {}).items():
            all_routes[route["routespec"]] = {
                "routespec": route["routespec"],
                "data": route.get("data", {}),
                "target": route["target"],
            }
        return all_routes



from traitlets.config import Config

from jupyterhub_traefik_proxy.consul import TraefikConsulProxy
from jupyterhub_traefik_proxy.etcd import TraefikEtcdProxy
from jupyterhub_traefik_proxy.toml import TraefikTomlProxy


def test_toml_deprecation(caplog):
    cfg = Config()
    cfg.TraefikTomlProxy.toml_static_config_file = 'deprecated-static.toml'
    cfg.TraefikTomlProxy.toml_dynamic_config_file = 'deprecated-dynamic.toml'
    p = TraefikTomlProxy(config=cfg)
    assert p.static_config_file == 'deprecated-static.toml'

    assert p.dynamic_config_file == 'deprecated-dynamic.toml'

    log = '\n'.join([record.msg for record in caplog.records])
    assert 'TraefikFileProvider.dynamic_config_file instead' in log
    assert 'static_config_file instead' in log


def test_etcd_deprecation(caplog):
    cfg = Config()
    cfg.TraefikEtcdProxy.kv_url = "http://1.2.3.4:12345"
    cfg.TraefikEtcdProxy.kv_username = "user"
    cfg.TraefikEtcdProxy.kv_password = "pass"

    p = TraefikEtcdProxy(config=cfg)
    assert p.etcd_url == "http://1.2.3.4:12345"
    assert p.etcd_username == "user"
    assert p.etcd_password == "pass"

    log = '\n'.join([record.msg for record in caplog.records])
    assert 'TraefikEtcdProxy.etcd_url instead' in log
    assert 'TraefikEtcdProxy.etcd_username instead' in log
    assert 'TraefikEtcdProxy.etcd_password instead' in log


def test_consul_deprecation(caplog):
    cfg = Config()
    cfg.TraefikConsulProxy.kv_url = "http://1.2.3.4:12345"
    cfg.TraefikConsulProxy.kv_username = "user"
    cfg.TraefikConsulProxy.kv_password = "pass"

    p = TraefikConsulProxy(config=cfg)
    assert p.consul_url == "http://1.2.3.4:12345"
    assert p.consul_username == "user"
    assert p.consul_password == "pass"

    log = '\n'.join([record.msg for record in caplog.records])
    assert 'TraefikConsulProxy.consul_url instead' in log
    assert 'TraefikConsulProxy.consul_username instead' in log
    assert 'TraefikConsulProxy.consul_password instead' in log



import argparse
import hashlib
import os
import platform
import sys
import tarfile
import textwrap
import warnings
import zipfile
from urllib.request import HTTPError, urlopen, urlretrieve

machine_map = {
    "x86_64": "amd64",
}


def checksum_file(path):
    """Compute the sha256 checksum of a path"""
    hasher = hashlib.sha256()
    with open(path, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            hasher.update(chunk)
    return hasher.hexdigest()


def fetch_checksums(traefik_version):
    """Fetch the checksum file from a traefik release"""
    url = (
        "https://github.com/traefik/traefik/releases"
        f"/download/v{traefik_version}/traefik_v{traefik_version}_checksums.txt"
    )
    checksums = {}
    print(f"Fetching checksums from {url}")
    try:
        urlopen(url)
    except HTTPError as e:
        print(f"Failed to retrieve checksum file: {e}")
        return {}
    with urlopen(url) as f:
        for line in f:
            line = line.decode("utf8", "replace").strip()
            if line.startswith("#") or not line:
                continue
            checksum, name = line.split()
            checksums[name] = checksum
    return checksums


def install_traefik(prefix, plat, traefik_version):
    plat = plat.replace("-", "_")
    if "windows" in plat:
        traefik_archive_extension = "zip"
        traefik_bin = os.path.join(prefix, "traefik.exe")
    else:
        traefik_archive_extension = "tar.gz"
        traefik_bin = os.path.join(prefix, "traefik")

    traefik_archive = (
        "traefik_v" + traefik_version + "_" + plat + "." + traefik_archive_extension
    )
    traefik_archive_path = os.path.join(prefix, traefik_archive)

    traefik_url = (
        "https://github.com/traefik/traefik/releases"
        f"/download/v{traefik_version}/{traefik_archive}"
    )

    if os.path.exists(traefik_bin):
        print(f"Traefik already exists at {traefik_bin}. Remove it to re-install.")
        print("--- Done ---")
        return

    print(f"Downloading traefik {traefik_version} from {traefik_url}...")
    urlretrieve(traefik_url, traefik_archive_path)

    checksums = fetch_checksums(traefik_version)
    expected_checksum = checksums.get(traefik_archive, None)

    if expected_checksum is not None:
        checksum = checksum_file(traefik_archive_path)
        if checksum != expected_checksum:
            raise OSError(f"Checksum failed {checksum} != {expected_checksum}")
    else:
        warnings.warn(
            f"Traefik {traefik_version} checksum could not be verified!",
            stacklevel=2,
        )

    print("Extracting the archive...")
    if traefik_archive_extension == "tar.gz":
        with tarfile.open(traefik_archive_path, "r") as tar_ref:
            tar_ref.extract("traefik", prefix)
    else:
        with zipfile.ZipFile(traefik_archive_path, "r") as zip_ref:
            zip_ref.extract("traefik.exe", prefix)
    os.chmod(traefik_bin, 0o755)
    print(f"Installed {traefik_bin}")
    os.unlink(traefik_archive_path)
    print("--- Done ---")


def main():
    # extract supported and default versions from urls
    parser = argparse.ArgumentParser(
        description="Dependency installer helper",
        formatter_class=argparse.RawTextHelpFormatter,
    )

    parser.add_argument(
        "--output",
        dest="installation_dir",
        default="./dependencies",
        help=textwrap.dedent(
            """\
            The installation directory (absolute or relative path).
            If it doesn't exist, it will be created.
            If no directory is provided, it defaults to:
            --- %(default)s ---
            """
        ),
    )

    machine = platform.machine()
    machine = machine_map.get(machine, machine)
    default_platform = f"{sys.platform}-{machine}"

    parser.add_argument(
        "--platform",
        dest="plat",
        default=default_platform,
        help=textwrap.dedent(
            """\
            The platform to download for.
            If no platform is provided, it defaults to:
            --- %(default)s ---
            """
        ),
    )

    parser.add_argument(
        "--traefik",
        action="store_true",
        help="DEPRECATED, IGNORED",
    )

    parser.add_argument(
        "--traefik-version",
        dest="traefik_version",
        default="2.9.8",
        help=textwrap.dedent(
            """\
            The version of traefik to download.
            If no version is provided, it defaults to:
            --- %(default)s ---
            """
        ),
    )
    if "--etcd" in sys.argv:
        sys.exit(
            "Installing etcd is no longer supported. Visit https://github.com/etcd-io/etcd/releases/"
        )
    if "--consul" in sys.argv:
        sys.exit(
            "Installing consul is no longer supported. Visit https://developer.hashicorp.com/consul/downloads"
        )

    args = parser.parse_args()
    deps_dir = args.installation_dir
    plat = args.plat
    traefik_version = args.traefik_version.lstrip("v")

    if args.traefik:
        print(
            "Specifying --traefik is deprecated and ignored. Only installing traefik is supported.",
            file=sys.stderr,
        )

    if os.path.exists(deps_dir):
        print(f"Using existing output directory {deps_dir}...")
    else:
        print(f"Creating output directory {deps_dir}...")
        os.makedirs(deps_dir)

    install_traefik(deps_dir, plat, traefik_version)


if __name__ == "__main__":
    main()



"""Traefik implementation

Custom proxy implementations can subclass :class:`Proxy`
and register in JupyterHub config:

.. sourcecode:: python

    from mymodule import MyProxy
    c.JupyterHub.proxy_class = MyProxy

Route Specification:

- A routespec is a URL prefix ([host]/path/), e.g.
  'host.tld/path/' for host-based routing or '/path/' for default routing.
- Route paths should be normalized to always start and end with '/'
"""

# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

import asyncio
import os
from itertools import chain

from traitlets import Any, Unicode, default, observe

from . import traefik_utils
from .proxy import TraefikProxy


class TraefikFileProviderProxy(TraefikProxy):
    """JupyterHub Proxy implementation using traefik and toml or yaml config file"""

    provider_name = "file"
    mutex = Any()

    @default("mutex")
    def _default_mutex(self):
        return asyncio.Lock()

    dynamic_config_file = Unicode(
        "rules.toml", config=True, help="""traefik's dynamic configuration file"""
    )

    dynamic_config_handler = Any()

    @default("dynamic_config_handler")
    def _default_handler(self):
        return traefik_utils.TraefikConfigFileHandler(self.dynamic_config_file)

    # If dynamic_config_file is changed, then update the dynamic config file handler
    @observe("dynamic_config_file")
    def _set_dynamic_config_file(self, change):
        self.dynamic_config_handler = traefik_utils.TraefikConfigFileHandler(
            self.dynamic_config_file
        )

    @default("dynamic_config")
    def _load_dynamic_config(self):
        try:
            # Load initial dynamic config from disk
            dynamic_config = self.dynamic_config_handler.load()
        except FileNotFoundError:
            dynamic_config = {}

        # fill in default keys
        # use setdefault to ensure these are always fully defined
        # and never _partially_ defined
        http = dynamic_config.setdefault("http", {})
        http.setdefault("services", {})
        http.setdefault("routers", {})
        jupyterhub = dynamic_config.setdefault("jupyterhub", {})
        jupyterhub.setdefault("routes", {})
        return dynamic_config

    def _persist_dynamic_config(self):
        """Save the dynamic config file with the current dynamic_config"""
        # avoid writing empty dicts, which traefik doesn't handle for some reason
        dynamic_config = self.dynamic_config
        if (
            not dynamic_config["http"]["routers"]
            or not dynamic_config["http"]["services"]
        ):
            # traefik can't handle empty dicts, so don't persist them.
            # But don't _remove_ them from our own config
            # I think this is a bug in traefik - empty dicts satisfy the spec
            # use shallow copy, which is cheap but most be done at every level where we modify keys
            dynamic_config = dynamic_config.copy()
            dynamic_config["http"] = http = dynamic_config["http"].copy()
            for key in ("routers", "services"):
                if not http[key]:
                    http.pop(key)
            if not http:
                dynamic_config.pop("http")
        self.dynamic_config_handler.atomic_dump(dynamic_config)

    async def _setup_traefik_dynamic_config(self):
        self.log.info(
            f"Creating the dynamic configuration file: {self.dynamic_config_file}"
        )
        await super()._setup_traefik_dynamic_config()

    async def _setup_traefik_static_config(self):
        self.static_config["providers"] = {
            "file": {"filename": self.dynamic_config_file, "watch": True}
        }
        await super()._setup_traefik_static_config()

    def _cleanup(self):
        """Cleanup dynamic config file as well"""
        super()._cleanup()
        try:
            os.remove(self.dynamic_config_file)
        except Exception as e:
            self.log.error(
                f"Failed to remove traefik configuration file {self.dynamic_config_file}: {e}"
            )

    async def _get_jupyterhub_dynamic_config(self):
        return self.dynamic_config["jupyterhub"]

    async def _apply_dynamic_config(self, traefik_config, jupyterhub_config=None):
        dynamic_config = {}
        dynamic_config.update(traefik_config)
        # file provider stores jupyterhub info in "jupyterhub" key
        if jupyterhub_config is not None:
            dynamic_config["jupyterhub"] = jupyterhub_config
        async with self.mutex:
            self.dynamic_config = traefik_utils.deep_merge(
                self.dynamic_config, dynamic_config
            )
            self._persist_dynamic_config()

    async def _delete_dynamic_config(self, traefik_keys, jupyterhub_keys):
        """Delete keys from dynamic configuration

        jupyterhub dynamic config is _inside_ traefik dynamic config,
        under the 'jupyterhub' key
        """
        jupyterhub_keys = (["jupyterhub"] + key_path for key_path in jupyterhub_keys)
        async with self.mutex:
            for key_path in chain(traefik_keys, jupyterhub_keys):
                parent = self.dynamic_config
                for key in key_path[:-1]:
                    if key in parent:
                        parent = parent[key]
                    else:
                        parent = {}
                        break

                # final key, time to delete
                key = key_path[-1]
                if key in parent:
                    parent.pop(key)
                else:
                    self.log.warning(
                        f"Missing dynamic config, nothing to delete: {'.'.join(key_path)}"
                    )

            self._persist_dynamic_config()

    async def get_route(self, routespec):
        """Return the route info for a given routespec.

        Args:
            routespec (str):
                A URI that was used to add this route,
                e.g. `host.tld/path/`

        Returns:
            result (dict):
                dict with the following keys::

                'routespec': The normalized route specification passed in to add_route
                    ([host]/path/)
                'target': The target host for this route (proto://host)
                'data': The arbitrary data dict that was passed in by JupyterHub when adding this
                        route.

            None: if there are no routes matching the given routespec
        """
        routespec = self.validate_routespec(routespec)
        router_alias = traefik_utils.generate_alias(routespec, "router")
        async with self.mutex:
            route = self.dynamic_config["jupyterhub"]["routes"].get(router_alias)
            if not route:
                return None
            return {
                "routespec": route["routespec"],
                "data": route["data"],
                "target": route["target"],
            }



from secrets import token_hex

from starlette.applications import Starlette
from starlette.responses import PlainTextResponse
from starlette.routing import Route, WebSocketRoute

messages = {
    "small": "",
    "large": token_hex(500_000),  # 1MiB
}

ws_chunk_size = 100_000


async def ws(websocket):
    await websocket.accept()
    size = websocket.path_params["size"]
    message = messages[size]
    await websocket.send_text(message)
    # for offset in range(0, len(message), ws_chunk_size):
    #     chunk = message[offset : offset + ws_chunk_size]
    #     await websocket.send_text(chunk)
    await websocket.send_text("")
    await websocket.close()


async def echo(request):
    size = request.headers["Request-Size"]
    return PlainTextResponse(messages[size])


async def index(request):
    return PlainTextResponse('ok')


routes = [
    Route("/", endpoint=index),
    WebSocketRoute(r"/{path:path}/ws/{size}", endpoint=ws),
    # WebSocketRoute(r"/{path:path}/ws/large", endpoint=ws_large),
    Route("/{path:path}", endpoint=echo),
]

app = Starlette(routes=routes)


if __name__ == "__main__":
    import uvicorn

    uvicorn.run(app, port=9000, log_level="info")



"""Install jupyterhub-traefik-proxy

Usage:

    pip install [-e] .
"""

import os
import sys

from setuptools import find_packages, setup
from setuptools.command.bdist_egg import bdist_egg

# ensure cwd is on sys.path
# workaround bug in pip 19.0
here = os.path.dirname(__file__)
if here not in sys.path:
    sys.path.insert(0, here)

import versioneer

cmdclass = versioneer.get_cmdclass()


class bdist_egg_disabled(bdist_egg):
    """Disabled version of bdist_egg

    Prevents setup.py install from performing setuptools' default easy_install,
    which it should never ever do.
    """

    def run(self):
        sys.exit(
            "Aborting implicit building of eggs. Use `pip install .` to install from source."
        )


if "bdist_egg" not in sys.argv:
    cmdclass["bdist_egg"] = bdist_egg_disabled

with open("README.md", encoding="utf8") as f:
    readme = f.read()

setup(
    name="jupyterhub-traefik-proxy",
    version=versioneer.get_version(),
    install_requires=open("requirements.txt").read().splitlines(),
    python_requires=">=3.6",
    author="Project Jupyter Contributors",
    author_email="jupyter@googlegroups.com",
    url="https://jupyterhub-traefik-proxy.readthedocs.io",
    project_urls={
        "Documentation": "https://jupyterhub-traefik-proxy.readthedocs.io",
        "Source": "https://github.com/jupyterhub/traefik-proxy/",
        "Tracker": "https://github.com/jupyter/traefik-proxy/issues",
    },
    # this should be a whitespace separated string of keywords, not a list
    keywords="jupyter jupyterhub traefik proxy",
    description="JupyterHub proxy implementation with traefik",
    long_description=readme,
    long_description_content_type="text/markdown",
    license="BSD",
    classifiers=[
        "Intended Audience :: Developers",
        "Intended Audience :: System Administrators",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
    ],
    packages=find_packages(),
    include_package_data=True,
    cmdclass=cmdclass,
    entry_points={
        "jupyterhub.proxies": [
            "traefik_consul = jupyterhub_traefik_proxy.consul:TraefikConsulProxy",
            "traefik_etcd = jupyterhub_traefik_proxy.etcd:TraefikEtcdProxy",
            "traefik_file = jupyterhub_traefik_proxy.fileprovider:TraefikFileProviderProxy",
            "traefik_toml = jupyterhub_traefik_proxy.toml:TraefikTomlProxy",
        ]
    },
)



"""Traefik implementation

Custom proxy implementations can subclass :class:`Proxy`
and register in JupyterHub config:

.. sourcecode:: python

    from mymodule import MyProxy
    c.JupyterHub.proxy_class = MyProxy

Route Specification:

- A routespec is a URL prefix ([host]/path/), e.g.
  'host.tld/path/' for host-based routing or '/path/' for default routing.
- Route paths should be normalized to always start and end with '/'
"""

# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

import base64
import string
from urllib.parse import urlparse

from traitlets import Any, Unicode, default

from .kv_proxy import TKvProxy


class TraefikConsulProxy(TKvProxy):
    """JupyterHub Proxy implementation using traefik and Consul"""

    provider_name = "consul"

    # Consul doesn't accept keys containing // or starting with / so we have to escape them
    key_safe_chars = string.ascii_letters + string.digits + "!@#$%^&*();<>-.+?:"

    consul_client_ca_cert = Unicode(
        config=True,
        allow_none=True,
        default_value=None,
        help="""Consul client root certificates""",
    )

    consul_url = Unicode(
        "http://127.0.0.1:8500",
        config=True,
        help="URL for the consul endpoint.",
    )
    consul_username = Unicode(
        "",
        config=True,
        help="Usrname for accessing consul.",
    )
    consul_password = Unicode(
        "",
        config=True,
        help="Password or token for accessing consul.",
    )

    kv_url = Unicode("DEPRECATED", config=True).tag(
        deprecated_in="1.0",
        deprecated_for="consul_url",
    )
    kv_username = Unicode("DEPRECATED", config=True).tag(
        deprecated_in="1.0",
        deprecated_for="consul_username",
    )
    kv_password = Unicode("DEPRECATED", config=True).tag(
        deprecated_in="1.0",
        deprecated_for="consul_password",
    )

    consul = Any()

    @default("consul")
    def _default_client(self):
        try:
            import consul.aio
        except ImportError:
            raise ImportError(
                "Please install python-consul2 package to use traefik-proxy with consul"
            )
        consul_service = urlparse(self.consul_url)
        kwargs = {
            "host": consul_service.hostname,
            "port": consul_service.port,
            "cert": self.consul_client_ca_cert,
        }
        if self.consul_password:
            kwargs.update({"token": self.consul_password})
        return consul.aio.Consul(**kwargs)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.log.warning(
            "Using traefik with consul is deprecated in jupyterhub-traefik-proxy 1.0 due to lack of support for the python-consul2 API client. Use etcd instead."
        )

    def _setup_traefik_static_config(self):
        provider_config = {
            "consul": {
                "rootKey": self.kv_traefik_prefix,
                "endpoints": [urlparse(self.consul_url).netloc],
            }
        }

        # FIXME: Same with the tls info
        if self.consul_client_ca_cert:
            provider_config["consul"]["tls"] = {"ca": self.consul_client_ca_cert}

        self.static_config.update({"providers": provider_config})
        return super()._setup_traefik_static_config()

    def _start_traefik(self):
        if self.consul_password:
            if self.consul_username:
                self.traefik_env.setdefault(
                    "CONSUL_HTTP_AUTH", f"{self.consul_username}:{self.consul_password}"
                )
            else:
                self.traefik_env.setdefault("CONSUL_HTTP_TOKEN", self.consul_password)
        super()._start_traefik()

    async def _kv_atomic_set(self, to_set):
        payload = []

        def append_payload(key, val):
            payload.append(
                {
                    "KV": {
                        "Verb": "set",
                        "Key": key,
                        "Value": base64.b64encode(val.encode()).decode(),
                    }
                }
            )

        for k, v in to_set.items():
            append_payload(k, v)

        try:
            await self.consul.txn.put(payload=payload)
        except Exception:
            self.log.exception("Error uploading payload to KV store!")
        else:
            self.log.debug("Successfully uploaded payload to KV store")

    async def _kv_atomic_delete(self, *to_delete):
        payload = []

        for key in to_delete:
            if key.endswith(self.kv_separator):
                verb = "delete-tree"
            else:
                verb = "delete"
            payload.append(
                {
                    "KV": {"Verb": verb, "Key": key},
                }
            )
        status, response = await self.consul.txn.put(payload=payload)
        # check response?

    async def _kv_get_tree(self, prefix):
        response = await self.consul.txn.put(
            payload=[
                {
                    "KV": {
                        "Verb": "get-tree",
                        "Key": prefix,
                    }
                }
            ]
        )
        kv_list = [
            (
                item["KV"]["Key"],
                base64.b64decode(item["KV"]["Value"] or '').decode("utf8"),
            )
            for item in response["Results"]
        ]
        return self.unflatten_dict_from_kv(kv_list, root_key=prefix)



import asyncio
import csv
import os
import sys
from concurrent.futures import ProcessPoolExecutor
from functools import partial
from pathlib import Path
from urllib.parse import urlparse

import requests
import websockets

performance_dir = Path(__file__).parent.resolve()

sys.path.insert(0, performance_dir)
import perf_utils


async def add_route_perf(proxy, route_idx, stdout_print):
    """
    Computes time taken (ns) to add the "route_idx"
    route to the proxy's routing table (which
    contains route_idx - 1 routes when ran sequentially).

    Returns a tuple:
    [
        route_idx(int):   the index of the route added
        time_taken(dict): the time it took to add the route
                          number "route_idx" to the proxy's
                          routing table.
                          keys:
                             'cpu': CPU time, 'time.process_time()'
                             'real': Real time, 'time.perf_counter()'
    ]
    """
    target = "http://127.0.0.1:9000"
    data = {"test": "test_" + str(route_idx), "user": "user_" + str(route_idx)}
    routespec = "/route/" + str(route_idx) + "/"
    time_taken = {}

    with perf_utils.measure_time(
        f"adding route_{route_idx}, took", stdout_print, time_taken
    ):
        await proxy.add_route(routespec, target, data)

    return route_idx, time_taken


async def delete_route_perf(proxy, route_idx, stdout_print):
    """
    Computes time taken (ns) to delete the "route_idx"
    route from the proxy's routing table (which
    contains route_idx + 1 routes when ran sequentially).

    It assumes the route to be deleted exists.

    Returns a tuple:
    [
        route_idx(int):   the index of the route deleted
        time_taken(dict): the time it took to delete the route
                          number "route_idx" from the proxy's
                          routing table.
                          keys:
                             'cpu': CPU time, 'time.process_time()'
                             'real': Real time, 'time.perf_counter()'
    ]
    """
    routespec = "/route/" + str(route_idx) + "/"
    time_taken = {}

    with perf_utils.measure_time(
        f"deleting route_{route_idx}, took", stdout_print, time_taken
    ):
        await proxy.delete_route(routespec)

    return route_idx, time_taken


async def get_all_routes_perf(proxy, iteration, stdout_print):
    """
    Computes time taken (ns) to retrieve all the
    routes from the proxy's routing table.

    Returns a tuple:
    [
        iteration(int):   the iteration index
        time_taken(dict): the time it took to get all the routes
                          keys:
                             'cpu': CPU time, 'time.process_time()'
                             'real': Real time, 'time.perf_counter()'
    ]
    """
    time_taken = {}
    with perf_utils.measure_time(
        f"getting all routes iteration_{iteration} took", stdout_print, time_taken
    ):
        await proxy.get_all_routes()

    return iteration, time_taken


async def run_methods_concurrent(method, proxy, routes, stdout_print, concurrency):
    semaphore = asyncio.BoundedSemaphore(concurrency)

    async def run_one(route_idx):
        # measurement: include wait for semaphore?
        # outer_time = {}
        # with perf_utils.measure_time(
        #     f"Awaiting {method.__name__} for route {route_idx}, took",
        #     stdout_print,
        #     outer_time,
        # ):
        async with semaphore:
            _, time_taken = await method(proxy, route_idx, stdout_print)
            # time_taken["full"] = outer_time["real"]
        return route_idx, time_taken

    tasks = [asyncio.ensure_future(run_one(route_idx)) for route_idx in range(routes)]
    with perf_utils.measure_time(
        f"Running {method.__name__} for {routes} total routes, took", True, {}
    ):
        _done, _running = await asyncio.wait(tasks, return_when=asyncio.ALL_COMPLETED)
    results = {}
    for route_idx, times in await asyncio.gather(*tasks):
        results[route_idx] = times
    return results


async def measure_methods_performance(
    concurrency,
    proxy_class,
    routes,
    stdout_print=True,
):
    async with perf_utils.get_proxy(proxy_class) as proxy:
        run = partial(
            run_methods_concurrent,
            concurrency=concurrency,
            proxy=proxy,
            routes=routes,
            stdout_print=stdout_print,
        )

        result = {}
        result["add"] = await run(add_route_perf)
        result["get_all"] = await run(get_all_routes_perf)
        result["delete"] = await run(delete_route_perf)

    return result


def make_http_req(public_url, routespec, request_size):
    """request functions run in the background via process pool"""
    req_url = public_url + routespec
    r = requests.get(req_url, headers={"Request-Size": request_size})
    r.raise_for_status()


def make_ws_req(public_url, routespec, request_size):
    """request functions run in the background via process pool"""
    netloc = urlparse(public_url).netloc
    req_url = f"ws://{netloc}{routespec}ws/{request_size}"

    async def f():
        async with websockets.connect(req_url) as websocket:
            msg = await websocket.recv()
            while msg != "":
                msg = await websocket.recv()

    asyncio.run(f())


async def measure_proxy_throughput(
    proxy_class,
    total_requests,
    concurrent_no,
    proto,
    request_size,
    backend_port,
    stdout_print=True,
):
    """
    Makes 'total_requests' GET http/websocket requests
    to the proxy with max 'concurrent_no' concurrent.

    Returns the throughput(number of requests/milisec)
    """
    pool = ProcessPoolExecutor(concurrent_no)

    async with perf_utils.get_proxy(proxy_class) as proxy:
        routespec = "/some_routespec/"
        target = "http://127.0.0.1:" + str(backend_port)
        data = {"test": "test1", "user": "username"}
        await proxy.add_route(routespec, target, data)

        asyncio.get_running_loop()

        time_taken = {}

        with perf_utils.measure_time(
            f"The {total_requests} {request_size} {proto} reqests with {concurrent_no} running concurrently, took",
            stdout_print,
            time_taken,
        ):
            if proto == "http":
                make_req = make_http_req
            else:
                make_req = make_ws_req
            tasks = [
                asyncio.wrap_future(
                    pool.submit(make_req, proxy.public_url, routespec, request_size)
                )
                for i in range(total_requests)
            ]
            await asyncio.gather(*tasks)

        real_time = time_taken["real"]

        throughput = total_requests / real_time
        print(
            f"{concurrent_no} concurrent requests: throughput = {throughput:.3f} requests/s\n"
        )

    return throughput


async def main():
    parser = perf_utils.configure_argument_parser()

    args = parser.parse_args()
    metric = args.metric
    concurrency = int(args.concurrency)
    proxy_class = args.proxy_class
    routes = int(args.routes)
    total_requests = int(args.total_requests)
    csv_filename = args.csv_filename
    test_iterations = int(args.test_iterations)
    print(args)

    loop = asyncio.get_running_loop()
    if not csv:
        loop.set_debug(True)  # Enable debug if we're just printing the results

    if metric == "methods":
        results = {}
        for i in range(test_iterations):
            print(
                f"Starting {metric} {concurrency=} measurement number {i} for {proxy_class} ...\n"
            )
            results[i] = await measure_methods_performance(
                concurrency, proxy_class, routes, csv_filename is None
            )

        if csv_filename:
            with open(csv_filename, mode="a+") as csv_file:
                samples = perf_utils.logspace_samples(routes)
                const_fields = {
                    "proxy": proxy_class,
                    "concurrency": concurrency,
                    "total_routes": routes,
                }
                fieldnames = list(const_fields.keys()) + [
                    "test_id",
                    "method",
                    "route_idx",
                    "cpu_time",
                    "real_time",
                ]
                writer = csv.DictWriter(csv_file, fieldnames=fieldnames)
                if os.stat(csv_filename).st_size == 0:
                    writer.writeheader()

                perf_utils.persist_methods_results_to_csv(
                    writer,
                    results,
                    test_iterations,
                    samples,
                    const_fields,
                )
    else:
        print(f"Started measuring {metric}")
        print(
            f"Running {total_requests} requests with up to {concurrency} concurrent...\n"
        )
        results = {}
        kind, _, size = metric.split("_")

        async with perf_utils.backend(concurrency) as backend_port:
            for i in range(test_iterations):
                results[i] = await measure_proxy_throughput(
                    proxy_class,
                    total_requests,
                    concurrency,
                    kind,
                    size,
                    backend_port,
                    not csv_filename,
                )

            print(f"Request throughput {size} {kind} requests: {results}")

        if csv_filename:
            with open(csv_filename, mode="a+") as csv_file:
                fieldnames = [
                    "test_id",
                    "throughput",
                ]
                const_fields = {
                    "proxy": proxy_class,
                    "kind": kind,
                    "size": size,
                    "total_requests": total_requests,
                    "concurrency": concurrency,
                }
                fieldnames.extend(const_fields.keys())
                writer = csv.DictWriter(csv_file, fieldnames=fieldnames)
                if os.stat(csv_filename).st_size == 0:
                    writer.writeheader()
                for i in range(test_iterations):
                    row = {}
                    row.update(const_fields)
                    row["test_id"] = i
                    row["throughput"] = results[i]
                    writer.writerow(row)


if __name__ == "__main__":
    asyncio.run(main())



import os
import string
from contextlib import contextmanager
from tempfile import NamedTemporaryFile
from urllib.parse import unquote

import escapism
from traitlets import Unicode


class KVStorePrefix(Unicode):
    def validate(self, obj, value):
        u = super().validate(obj, value)
        if u.endswith("/"):
            u = u.rstrip("/")

        proxy_class = type(obj).__name__
        if "Consul" in proxy_class and u.startswith("/"):
            u = u[1:]

        return u


def generate_rule(routespec):
    routespec = unquote(routespec)
    if routespec.startswith("/"):
        # Path-based route, e.g. /proxy/path/
        rule = f"PathPrefix(`{routespec}`)"
    else:
        # Host-based routing, e.g. host.tld/proxy/path/
        host, path_prefix = routespec.split("/", 1)
        rule = f"Host(`{host}`) && PathPrefix(`/{path_prefix}`)"
    return rule


_safe = set(string.ascii_letters + string.digits + "-")


def generate_alias(routespec, kind=""):
    """Generate an alias for a routespec

    A safe string for use in key-value store keys, etc.
    """
    alias = escapism.escape(routespec, safe=_safe)
    if kind:
        alias = f"{kind}_{alias}"
    return alias


# atomic writing adapted from jupyter/notebook 5.7
# unlike atomic writing there, which writes the canonical path
# and only use the temp file for recovery,
# we write the temp file and then replace the canonical path
# to ensure that traefik never reads a partial file


@contextmanager
def atomic_writing(path):
    """Write temp file before copying it into place

    Avoids a partial file ever being present in `path`,
    which could cause traefik to load a partial routing table.
    """
    fileobj = NamedTemporaryFile(
        prefix=os.path.abspath(path) + "-tmp-", mode="w", delete=False
    )
    try:
        with fileobj as f:
            yield f
        os.replace(fileobj.name, path)
    finally:
        try:
            os.unlink(fileobj.name)
        except FileNotFoundError:
            # already deleted by os.replace above
            pass


class TraefikConfigFileHandler:
    """Handles reading and writing Traefik config files. Can operate
    on both toml and yaml files"""

    def __init__(self, file_path):
        file_ext = file_path.rsplit('.', 1)[-1]
        if file_ext == 'yaml':
            try:
                from ruamel.yaml import YAML
            except ImportError:
                raise ImportError(
                    "jupyterhub-traefik-proxy requires ruamel.yaml to use YAML config files"
                )
            config_handler = YAML(typ="rt")
        elif file_ext == 'toml':
            import toml as config_handler
        else:
            raise TypeError("type should be either 'toml' or 'yaml'")

        self.file_path = file_path
        # Redefined to either yaml.dump or toml.dump
        self._dump = config_handler.dump
        # self._dumps = config_handler.dumps
        # Redefined by __init__, to either yaml.load or toml.load
        self._load = config_handler.load

    def load(self):
        """Depending on self.file_path, call either yaml.load or toml.load"""
        with open(self.file_path) as fd:
            return self._load(fd)

    def dump(self, data):
        with open(self.file_path, "w") as f:
            self._dump(data, f)

    def atomic_dump(self, data):
        """Save data to self.file_path after opening self.file_path with
        :func:`atomic_writing`"""
        with atomic_writing(self.file_path) as f:
            self._dump(data, f)


def deep_merge(a, b):
    """Merges dict b into dict a, returning a

    Note: a is modified in the process!
    """
    for k, v in b.items():
        if k in a:
            if isinstance(a[k], dict):
                # recursive update
                a[k] = deep_merge(a[k], v)
            else:
                a[k] = v
        else:
            a[k] = v
    return a
